//*************************************************************************************
//	Copyright 2000 Raymond A. Virzi, Jr.
//	You may freely use and distribute this source code provide the above copyright notice
//	and this statement remains intact within the source module or any modified versions of it.

///////////////////////////////////////////////////////////////
// Pattern Class: CObserver
// Derive from this class objects that depend on other objects via an 'update' function.
//
// CHANGE HISTORY:
// 05/03/00	- Initial Revision.
// 05/04/00 - Uploaded to Codeguru.
///////////////////////////////////////////////////////////////
#include <warnings.h>

#ifndef __OBSERVER_H
#define __OBSERVER_H

#include <crtdbg.h>
#define assert(a) _ASSERT(a)

#include <algorithm>
#include <utility>
#include <map>
#include <set>

// CObserver state flags
#define S_NEEDS_UPDATE	0x01	// object needs to be updated
#define S_NEEDS_CHECK	0x02	// object needs to check its subjects for update
#define S_ALLDIRTY		0x04	// all subjects are dirty
#define S_CHANGED		0x08	// true if a change occurred during OnUpdate()
// The following four flags are for cyclic operations
#define S_UPDATING		0x10	// true during OnUpdate() call
#define S_PROPAGATING	0x20	// true if last change is currently propagating
#define S_NOTIFYING		0x40	// true during notification phase
#define S_CHECKING		0x80	// true during check phase

// Macros for flag manipulation - these work within private scope of CObserver
#define set_flags(mask, value)	(value ? (_flags |= (mask)) : (_flags &= ~(mask)))
	// To set all bits in 'mask' to the same 'value'
#define get_flags(mask)			(_flags & (mask))
	// To get the bits specified in 'mask'. 
	// Use 'get_flags(mask) == mask' to test if all the flags are on
	// Use 'get_flags(mask) > 0' to test if at least one of the flags is on

class CObserver
{
	typedef std::set<CObserver*> plist_type;
	typedef std::map<CObserver*, bool> tlist_type;	// key = pointer to subject, val = dirty flag

	plist_type _plist;	// observers
	tlist_type _tlist;	// subjects
	unsigned char _flags;	// state flags

	void RemoveSubjectSafely(CObserver& subject)
	// Removes subject without modifying the subject being removed
	{
		// Disconnect subject from this observer
		tlist_type::iterator at = _tlist.find(&subject);
		if(at != _tlist.end())
			_tlist.erase(at);
		
		Invalidate();
	}
	void RemoveObserverSafely(CObserver& observer)
	// Removes observer without modifying the observer being removed
	{
		// Disconnect observer from this subject
		plist_type::iterator at = _plist.find(&observer);
		if(at != _plist.end())
			_plist.erase(at);
	}
	void SetDirtyFlag(CObserver& subject, bool value)
	{
		tlist_type::iterator at = _tlist.find(&subject);
		if(at != _tlist.end())
		{
			at->second = value;
			if(!value)
				set_flags(S_ALLDIRTY, false);
		}
	}
	void NotifyCheck(CObserver *from)		// call only when object is invalidated
	{
		if(!get_flags(S_CHECKING))
		{
			if(from)
				SetDirtyFlag(*from, true);

			if(!get_flags(S_NEEDS_CHECK))
			{
				set_flags(S_NEEDS_CHECK, true);
				if(!get_flags(S_NOTIFYING))		// save some time
				{
					plist_type::iterator pos = _plist.begin();
					set_flags(S_NOTIFYING, true);
					while(pos != _plist.end())
					{
						(*pos)->NotifyCheck(this);
						++pos;
					}
					set_flags(S_NOTIFYING, false);
				}
			}
		}
	}
	void Validate()		// clears all dirty flags
	{
		tlist_type::iterator tpos = _tlist.begin();
		while(tpos != _tlist.end())
		{
			tpos->second = false;
			++tpos;
		}
		set_flags(S_ALLDIRTY | S_NEEDS_UPDATE | S_NEEDS_CHECK, false);
	}
	void SendNotify()
	{
		// If observers call Notify() any other time than during an update,
		// it is a temporary perturbation and must be either be immediately
		// propagated or immediately erased by the next update.
		if(!get_flags(S_NEEDS_UPDATE | S_UPDATING) && IsObserver())
			set_flags(S_NEEDS_UPDATE, true);

		// Invalidate observers - must be done before propagation
		plist_type::iterator pos;
		set_flags(S_NOTIFYING, true);
		pos = _plist.begin();
		while(pos != _plist.end())
		{
			(*pos)->Invalidate(this);
			++pos;
		}
		set_flags(S_NOTIFYING, false);

		if(get_flags(S_PROPAGATING))
		{
			// update all observers immediately
			pos = _plist.begin();
			while(pos != _plist.end())
			{
				(*pos)->Update(true);	// continue depth-first propagation
				++pos;
			}
			if(!get_flags(S_UPDATING))
			{
				set_flags(S_PROPAGATING, false);	// set back to default - insurance only
				if(IsObserver())	// perturbation - re-notify after propagation
					Notify();
			}
		}
	}
protected:
	CObserver()
	{
		_flags = 0;
	}
	virtual ~CObserver()
	{
		// Remove this subject from all observers
		plist_type::iterator pos = _plist.begin();
		while(pos != _plist.end())
		{
			(*pos)->RemoveSubjectSafely(*this);
			++pos;
		}
		// Remove this observer from all subjects - do this after observers are gone
		RemoveAllSubjects();
	}
	CObserver(const CObserver& other)	// copy constructor - starts from scratch!
	{
		_flags = 0;
	}
	CObserver& operator=(const CObserver& other) { return *this;}	// nothing copied
	void AddSubject(CObserver& subject)
	// Registers a new subject with this observer.
	{
		// Connect observer and subject
		_tlist.insert(std::make_pair(&subject,false));
		subject._plist.insert(this);

		Invalidate();
	}
	void RemoveSubject(CObserver& subject)
	// Unregisters a subject that was previously registered.
	{
		// Disconnect this observer from subject
		subject.RemoveObserverSafely(*this);
		// Disconnect subject from this observer
		RemoveSubjectSafely(subject);
	}
	void RemoveAllSubjects()
	// Unregisters all subjects.
	{
		tlist_type::iterator tpos = _tlist.begin();
		while(tpos != _tlist.end())
		{
			tpos->first->RemoveObserverSafely(*this);
			++tpos;
		}
		_tlist.clear();
		Invalidate();
	}
	void AddSubjectsOf(CObserver& observer)
	// Registers all the subjects of another observer.
	{
		tlist_type::iterator tpos = observer._tlist.begin();
		while(tpos != observer._tlist.end())
		{
			AddSubject(*(tpos->first));
			++tpos;
		}
	}
	virtual void OnUpdate()	{}	// Virtual function to do the updating. Call Notify() if the value changes.
	void Update(bool propagate = false)
	// Updates the objects and, if its value has changed, invlidates all its observers.
	{
		if(get_flags(S_PROPAGATING))	// freeze during propagation phase
			return;
		if(get_flags(S_UPDATING | S_CHECKING))	// cyclic check or update - wait until next cycle
			return;

		while(get_flags(S_NEEDS_CHECK | S_NEEDS_UPDATE))
		{		
			if(get_flags(S_NEEDS_CHECK) && !get_flags(S_NEEDS_UPDATE))
			{
				set_flags(S_NEEDS_CHECK, false);	// validate here for cyclic safety

				// check the subjects
				set_flags(S_CHECKING, true);
				tlist_type::iterator tpos = _tlist.begin();
				while(tpos != _tlist.end())
				{
					if(tpos->second)	// subject is dirty - may need update
					{
						tpos->first->Update();
						if(get_flags(S_NEEDS_UPDATE))	// stop searching as soon as we know we need to udpate
							break;
					}
					++tpos;
				}
				set_flags(S_CHECKING, false);
			}

			if(get_flags(S_NEEDS_UPDATE))
			{
				// don't validate here - Notify() will handle
				set_flags(S_UPDATING, true);
				set_flags(S_PROPAGATING, propagate);
				set_flags(S_CHANGED, false);
				OnUpdate();		// the rest of the subjects will be checked here automatically
				Validate();		// clear dirty flags after update
				if(get_flags(S_CHANGED))
					SendNotify();		// this could invalidate us again (cyclic)
				set_flags(S_PROPAGATING | S_UPDATING, false);
			}
		}
	}
	bool IsDirty(CObserver& subject) const
	{
		tlist_type::const_iterator at = _tlist.find(&subject);
		if(at != _tlist.end())
		{
			return (get_flags(S_ALLDIRTY) ? true : at->second);
		}
		else
			return false;
	}
	void Notify(bool propagate)		// Notify with propagation control
	{
		bool oldprop = get_flags(S_PROPAGATING) > 0;
		set_flags(S_PROPAGATING, propagate);
		Notify();
		set_flags(S_PROPAGATING, oldprop);
	}
	void Notify()
	// Notifies all observers that this object has changed. Call only when
	// the object's value actually changes.
	{
		if(get_flags(S_UPDATING))
			set_flags(S_CHANGED, true);		// delayed notification during update
		else
			SendNotify();
	}
	void Invalidate(CObserver *from = NULL)
	// Invalidates the object and notifies all its observers.
	{
		if(!get_flags(S_UPDATING) || get_flags(S_NOTIFYING))	// can't invalidate during update unless a cycle occurred
		{
			if(from)
				SetDirtyFlag(*from, true);
			else
				set_flags(S_ALLDIRTY, true);

			if(!get_flags(S_NEEDS_UPDATE))
			{
				set_flags(S_NEEDS_UPDATE, true);
				NotifyCheck(NULL);
			}
		}
	}
public:
	bool IsObserver() const { return !_tlist.empty();}
	bool IsSubject() const { return !_plist.empty();}
	void Propagate() 
	{ 
		Notify(true);		// propagate current value
	}
};


#endif
