// Test program for observer pattern


#include <cstdio>
#include <ctime>
#include <list>
#include <vector>
#include "observer.h"

class Number : public CObserver
{
public:
	typedef int num_type;
	static char _next_id;
private:
	num_type _val;
	char _id;
	char next_id() const { return _next_id++;}
public:
	Number(num_type val = 0)
		: _val(val), _id(0)
	{}
	virtual ~Number() {}
	void generate_id() { _id = next_id();}
	char id() const { return _id;}
	Number& operator=(Number val)
	{
		printf("Updating number %c: ", id());
		SetValue(val._val);
		return *this;
	}
	operator num_type()
	{ 
		Update();		// for derived classes
		return _val;
	}
	bool SetValue(num_type val)
	{
		printf("%i -> %i\n", _val, val);
		if(!(_val == val))
		{
			_val = val;
			Notify();
			return true;
		}
		return false;
	}
	void Reset() { _val = 0;}
};


class Sum : public Number
{
	typedef std::list<Number*> nlist_type;

	nlist_type _numlist;
public:
	Sum(num_type val = 0)
		: Number(val)
	{}
	void OnUpdate()
	{
		if(_numlist.empty())	// empty list means this is subject only
			return;
		num_type _sum = 0;
		nlist_type::iterator pos = _numlist.begin();
		while(pos != _numlist.end())
		{
			_sum += (num_type)(**pos);		// subjects updated here
			++pos;
		}
		printf("Updating sum %c: ", id());
		SetValue(_sum);		// calls Notify() for us
	}
	void AddNumber(Number& pnum)
	{
		printf("Adding subject to %c -> %c\n", id(), pnum.id());
		_numlist.push_back(&pnum);
		AddSubject(pnum);
	}
	Sum& operator=(Sum val) 
	{ 
		Number::operator=(val);
		return *this;
	}
	void Reset()
	{
		_numlist.clear();
		RemoveAllSubjects();
		Number::Reset();
	}
};

class CopyCat : public Number
{
	Number *_num;		// object to mimick
public:
	CopyCat()
		: _num(NULL)
	{}
	CopyCat& operator=(num_type val)
	{
		Number::operator=(val);
		return *this;
	}
	void OnUpdate()
	{
		if(!_num)
			return;

		num_type val = (num_type)(*_num);		// subject is updated here
		printf("Copying %c to %c: ", _num->id(), id());
		SetValue(val);			// calls Notify() for us
	}
	void Watch(Number& pnum)
	{
		printf("Adding subject to %c -> %c\n", id(), pnum.id());
		_num = &pnum;
		AddSubject(pnum);
	}
};


#define VALRANGE 10
char Number::_next_id = 'A';

void main(int argc, char* argv[])
{
	//------------------------------------------------------
	// Hierarchical update test
	printf("Hierarchical update test:\n\n");

	Number::_next_id = 'A';
	int list_size = 50;
	std::vector<Number> nlist(list_size);	// list of 'list_size' random numbers
	std::vector<Sum> slist(10);		// list of 10 sums each summing 10 numbers chosen from nlist
	Sum sumall;						// sum of all 10 sums in slist

	// initialize
	int i, count, choice;
	srand(time(NULL));
	Number::_next_id = 'A';
	for(i = 0; i < list_size; ++i)
	{
		nlist[i].generate_id();
		nlist[i] = (int)((rand() / (double)RAND_MAX) * VALRANGE);
	}
	for(i = 0; i < 10; ++i)
	{
		slist[i].generate_id();
		for(count = 0; count < 10; ++count)
		{
			choice = (int)((rand() / (double)RAND_MAX) * list_size);
			slist[i].AddNumber(nlist[choice]);
		}
	}
	sumall.generate_id();
	for(i = 0; i < 10; ++i)
	{
		sumall.AddNumber(slist[i]);
	}

	Number::num_type total, newval;
	
	printf("Accessing overall sum\n");
	total = sumall;		// Forces update for internal printout
	printf("Beginning random assignments at lowest level\n");
	count = 50;
	while(count--)
	{
		// get new value
		newval = (Number::num_type)((rand() / (double)RAND_MAX) * VALRANGE);

		// choose a number to update
		choice = (int)((rand() / (double)RAND_MAX) * list_size);

		// These are the interesting lines to trace!
		nlist[choice] = newval;		// change the value
		total = sumall;			// read the overall sum to force update
	}

	printf("\n\n");

	//-----------------------------------------------
	// Cyclic update test
	printf("Cyclic update test:\n\n");

	list_size = 10;
	std::vector<CopyCat> clist(list_size);		// list of 'list_size' numbers connected in a cycle

	// initialize
	Number::_next_id = 'A';
	for(i=0; i < list_size; i++)
	{
		clist[i].generate_id();
	}
	for(i=0; i < list_size-1; i++)
	{
		clist[i].Watch(clist[i+1]);
	}
	// complete the cycle
	clist[list_size-1].Watch(clist[0]);

	count = 10;
	while(count--)
	{
		// get new value
		newval = (Number::num_type)((rand() / (double)RAND_MAX) * VALRANGE);

		// choose a number to update
		choice = (int)((rand() / (double)RAND_MAX) * list_size);

		// These are the interesting lines to trace!
		clist[choice] = newval;		// change the value
		total = clist[choice];		// read value to force update
	}

	printf("\n\n");

	//----------------------------------------------------------
	// Fibonacci sequence
	printf("Fibonacci sequence test:\n\n");

	// Set up our list of 10 sums as a fibonacci sequence
	sumall.Reset();
	for(i = 0; i < 10; ++i)
	{
		slist[i].Reset();
	}
	printf("Assigning to bottom values\n");
	slist[0] = 1;		// start the sequence
	slist[1] = 1;
	for(i = 2; i < 10; ++i)
	{
		slist[i].AddNumber(slist[i-1]);
		slist[i].AddNumber(slist[i-2]);
	}
	printf("Accessing highest number\n");
	total = slist[9];	// update the last number

	printf("\n");

	// Continue the sequence by shifting the top values to the bottom
	printf("Continuing sequence by shifting top two values to the bottom\n");
	slist[0] = slist[8];
	slist[1] = slist[9];
	printf("Accessing highest number\n");
	total = slist[9];

	printf("\n\n");

	printf("Immediate propagation test with Fibonacci sequence:\n\n");

	printf("Resetting bottom values\n");
	slist[0] = 1;
	slist[1] = 1;
	printf("Request to propagate immediately\n");
	slist[0].Propagate();

	printf("\n\n");

	printf("Perturbation at center of sequence:\n\n");

	printf("Setting fifth number to zero\n");
	slist[4] = 0;			// add perturbation
	printf("Request to propagate immediately\n");
	slist[4].Propagate();	// propagate it
	printf("Accessing highest number\n");
	total = slist[9];		// relax perturbation on next update

	printf("\n\n");
}
